# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import unicode_literals

import json

import six
from jinja2 import nodes
from jinja2.ext import Extension


def json_hard_update(base, update):
    result = json.loads(base)
    result.update(json.loads(update))
    return json.dumps(result, sort_keys=True, indent=4)


def _patch(base, patch):
    for key, value in six.iteritems(patch):
        if key in base and type(base[key]) == dict and type(value) == dict:
            _patch(base[key], value)
        else:
            base[key] = value


def json_soft_update(base, update):
    base_json = json.loads(base)
    update_json = json.loads(update)

    _patch(base_json, update_json)

    return json.dumps(base_json, sort_keys=True, indent=4)


class JSONOverrideExtension(Extension):
    """
    Adds an override keyword to the jinja syntax. This allow to update
    2 json dictionaries together.
    Syntax is as follow:

    {% override "template.json" %}
    {
      "attr1": "override"
    }
    {% endoverride %}

    with template.json being a valid template name.

    The override decorated dictionary will update the dictionary defined in the
    passed template
    """
    tags = set(['override'])

    def parse(self, parser):
        lineno = next(parser.stream).lineno

        filename = parser.stream.expect('string').value

        body = parser.parse_statements(['name:endoverride'], drop_needle=True)

        return nodes.CallBlock(
            self.call_method('_override', [nodes.Const(filename)]),
            [],
            [],
            body,
        ).set_lineno(lineno)

    def _override(self, filename, caller=None):
        override_str = caller()
        base_str = self.environment.get_template(filename).render()

        return json_hard_update(base_str, override_str)


class JSONPatchExtension(Extension):
    """
    Adds an patch keyword to the jinja syntax. This allow to update
    2 json dictionaries together.
    Syntax is as follow:

    {% patch "template.json" %}
    {
      "attr1": {
      "value': "override"
      }
    }
    {% endpatch %}

    with template.json being a valid template name.

    All the values that the decorated dictionary defines will be updated in the
    passed template, with respect to hierarchy, and leaving most values as is
    """
    tags = set(['patch'])

    def parse(self, parser):
        lineno = next(parser.stream).lineno

        filename = parser.stream.expect('string').value

        body = parser.parse_statements(['name:endpatch'], drop_needle=True)

        return nodes.CallBlock(
            self.call_method('_override', [nodes.Const(filename)]),
            [],
            [],
            body,
        ).set_lineno(lineno)

    def _override(self, filename, caller=None):
        override_str = caller()
        base_str = self.environment.get_template(filename).render()

        return json_soft_update(base_str, override_str)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import unicode_literals

import collections
import string
import warnings

import six
from six.moves.urllib.parse import urlencode

from pyramid_mock_server.util import make_operation_from_path


class QueryUrlFormatter(string.Formatter):
    """ Appends unused kwargs at the end as query_urls
    """

    def format(self, format_string, *args, **kwargs):
        formated_str = super(QueryUrlFormatter, self).format(format_string, *args, **kwargs)
        if self.unused_kwargs:
            return '?'.join((formated_str, urlencode(self.unused_kwargs)))

        return formated_str

    def check_unused_args(self, used_args, args, kwargs):
        self.unused_kwargs = collections.OrderedDict(sorted(kwargs.items()))
        for arg in used_args:
            self.unused_kwargs.pop(arg, None)


query_url_formatter = QueryUrlFormatter()


def get_all_mocks_operations(response_collection, resources):
    """ Return all the urls/http_operation pairs that the mocks are
    defining.
    This is useful for testing all the mocks against a spec for example.

    :param response_collection: ResponseCollection of all mocks
    :param resources: iterable of (path, method) pairs
    """

    operations = []

    for endpoint_path, endpoint_method in resources:
        operation = make_operation_from_path(endpoint_path, endpoint_method)
        variations = response_collection.get_variations(operation)
        replacement_dicts = variations.generate_string_replacement_dict()

        for replacement_dict in replacement_dicts:
            operations.append((
                query_url_formatter.format(
                    endpoint_path,
                    **replacement_dict
                ),
                endpoint_method,
            ))

    return operations


def get_swagger20_resources_iterator_from_pyramid_swagger(config):
    """ Extract the swagger2.0 defined path from a pyramid swagger configuration

    Only call this after `config.include('pyramid_swagger')`

    :param config: the pyramid config
    :return: an iterator over the swagger resources yielding (path, http_verb) tuples
    """
    swagger_schema = config.registry.settings.get('pyramid_swagger.schema20')
    if not swagger_schema:
        warnings.warn(
            'read_resources_from_pyramid_swagger is True but pyramid_swagger is not available'
        )
    else:
        base_path = swagger_schema.spec_dict.get('basePath', '').rstrip('/')
        for resource in six.itervalues(swagger_schema.resources):
            for op in six.itervalues(resource.operations):
                yield base_path + op.path_name, op.http_method



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import unicode_literals

import venusian
from pyramid.response import Response

from pyramid_mock_server.mock_loader import load_responses
from pyramid_mock_server.response_collection import ResponseCollection
from pyramid_mock_server.util import make_operation_from_path


def make_response(mock_response):
    """Default endpoint response w/ its corresponding sample json data."""
    return Response(
        mock_response.json_str,
        status=mock_response.http_response_code,
        content_type=str('application/json'),
        charset='UTF-8',
    )


def make_servlet_view_fn(
    endpoint_operation,
    response_collection,
):
    """Create the given endpoint's servlet view function.

    :param endpoint_operation: Operation that will spawn the endpoint
    :param response_collection: ResponseCollection instance.
    :return: a pyramid view function.
    """
    variations = response_collection.get_variations(endpoint_operation)
    if not variations:
        def no_response_found(request):
            raise LookupError(
                u"No json response found for {0} {1}".format(
                    endpoint_operation.response_name,
                    endpoint_operation.http_verb
                )
            )
        return no_response_found

    def custom_view(request):
        arg_list = variations.get_arg_list()
        arg_dict = dict(
            (arg, request.matchdict[arg])
            for arg in arg_list
            if arg in request.matchdict
        )
        arg_dict.update(dict(
            (arg, request.GET[arg]) for arg in arg_list
            for arg in arg_list
            if arg in request.GET
        ))

        matched_mock_response = variations.match_arg_dict(arg_dict)
        return make_response(matched_mock_response)

    return custom_view


def setup_routes_views(
    config,
    responses_path,
    resources,
    custom_view_packages=None,
    excluded_paths=None,
):
    """ Creates the default views when needed, get the custom views, register
    them all in a pyramid config.

    :param config: pyramid config.
    :param responses_path: file path to the root json responses directory.
    :param resources: iterable of (path, method) string tuples.
    :param custom_view_packages: array of packages object that we should load custom view from.
    :param excluded_paths: array of path for which this function should not try to
        register views. Note ['foo'] would also exclude 'foo/whatever'.
    """

    routes_added = set()
    routes_operation_added = set()
    view_registry = {}
    if not excluded_paths:
        excluded_paths = []
    if not custom_view_packages:
        custom_view_packages = []

    responses = load_responses(responses_path)
    response_collections = ResponseCollection(responses)

    for package in custom_view_packages:
        _collect_custom_views(package, view_registry)

    for endpoint_path, http_verb in resources:
        endpoint_operation = make_operation_from_path(endpoint_path, http_verb)

        # Ignore excluded paths/names
        if any(map(endpoint_path.startswith, excluded_paths)):
            continue

        # Only add a route once
        if endpoint_operation.response_name not in routes_added:
            config.add_route(endpoint_operation.response_name, endpoint_path)
            routes_added.add(endpoint_operation.response_name)

        # Only add view once, for each resource
        if endpoint_operation not in routes_operation_added:
            if endpoint_operation not in view_registry:
                view_registry[endpoint_operation] = make_servlet_view_fn(
                    endpoint_operation,
                    response_collections,
                )
            config.add_view(
                view_registry[endpoint_operation],
                route_name=endpoint_operation.response_name,
                request_method=endpoint_operation.http_verb,
            )
            routes_operation_added.add(endpoint_operation)


def register_custom_view(path, http_verb):
    """ Associate the decorated function with the given path.
    This allows for custom views not automatically generated to be used.

        :param path: path as defined in swagger spec
        :param http_verb: http verb ('PUT', 'GET', ...)  for this view
        :return: decorator for the view
    """

    def wrapper(wrapped):
        def callback(scanner, _, ob):
            scanner.view_registry[make_operation_from_path(path, http_verb)] = wrapped
        venusian.attach(wrapped, callback, category='pyramid_mock_server')
        return wrapped

    return wrapper


def _collect_custom_views(package, view_registry):
    """ Collect all the custom views marked with the register_custom_view
    decorator and add them to the view_registry

    :param package: package to look for decorated views into
    :param view_registry: dictionary containing the path-view function pairs
    """
    scanner = venusian.Scanner(view_registry=view_registry)
    scanner.scan(package, categories=('pyramid_mock_server',))
